# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .......schema_classes import ConditionClass
from .......schema_classes import ConjunctiveCriterionClass
from .......schema_classes import CriterionClass
from .......schema_classes import FilterClass


Condition = ConditionClass
ConjunctiveCriterion = ConjunctiveCriterionClass
Criterion = CriterionClass
Filter = FilterClass

# fmt: on
